#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <regexp.h>

#include <sqlite3/sqlite3.h>

#include "ocl.h"
#include "ocliterator.h"
#include "oclfile.h"
#include "ocldate.h"



struct OclDatasource {
  char* url;
  char* protocol;
  char* host;
  char* file;
  int port;
  char* name;
  char* passwd;
  char* schema;
  sqlite3* database;
};

struct SQLStatement {
  char* text; 
  struct OclDatasource* connection; 
  sqlite3_stmt* statement;
  struct OclIterator* resultSet;  
};

struct SQLStatement* createSQLStatement(void) { 
  struct SQLStatement* res = (struct SQLStatement*) malloc(sizeof(struct SQLStatement)); 
  res->text = ""; 
  res->connection = NULL; 
  res->statement = NULL; 
  res->resultSet = NULL; 
  return res;
} 

struct OclDatasource* createOclDatasource(void) 
{ struct OclDatasource* result = 
     (struct OclDatasource*) malloc(sizeof(struct OclDatasource));
  result->url = "";
  result->protocol = "";
  result->host = "";
  result->file = "";
  result->port = 0;
  result->name = "";
  result->passwd = "";
  result->schema = "";
  result->database = NULL; 
  return result; 
}


struct OclDatasource* getConnection(char* url, char* name, char* passwd)
{ sqlite3* db;
  char* errs;  
  int succ = sqlite3_open(url, &db);
 
  if (succ == SQLITE_OK)
  { 
    struct OclDatasource* res = createOclDatasource();
    res->url = url;
    res->name = name;
    res->passwd = passwd;
    res->database = db; 
    return res; 
  }

  return NULL; 
}


struct OclDatasource* newOclDatasource(void)
{
  struct OclDatasource* result = createOclDatasource(); 
  return result;
}


struct OclDatasource* newSocket(char* host, int port)
{
  struct OclDatasource* db = createOclDatasource();
  db->host = host;
  db->port = port;
  return db; 
}


struct OclDatasource* newURL(char* s)
{
  struct OclDatasource* db = createOclDatasource();
  db->url = s;
  return db; 
}


struct OclDatasource* newURL_PHF(char* p, char* h, char* f)
{
  struct OclDatasource* db = createOclDatasource();
  db->protocol = p;
  db->host = h;
  db->file = f;
  return db; 
}


struct OclDatasource* newURL_PHNF(char* p, char* h, int n, char* f)
{
  struct OclDatasource* db = createOclDatasource();
  db->protocol = p;
  db->host = h;
  db->port = n;
  db->file = f;
  return db; 
}


struct SQLStatement* createStatement_OclDatasource(struct OclDatasource* self)
{
  struct SQLStatement* ss = createSQLStatement();
  ss->text = "";
  ss->connection = self; 

  if (self->database != NULL) 
  { sqlite3_stmt* statement; 
    const void* rem; 
    sqlite3_prepare(self->database, ss->text, 0, &statement, &rem);
    ss->statement = statement;  
  } 

  return ss; 
}

struct SQLStatement* prepare_OclDatasource(struct OclDatasource* self, char* stat)
{
  struct SQLStatement* ss = createSQLStatement();
  ss->text = stat;
  ss->connection = self; 

  if (self->database != NULL) 
  { sqlite3_stmt* statement; 
    const void* rem; 
    sqlite3_prepare(self->database, ss->text, -1, &statement, &rem);
    ss->statement = statement;  
  } 

  return ss; 
}

struct SQLStatement* prepareStatement_OclDatasource(struct OclDatasource* self, char* stat)
{
  return prepare_OclDatasource(self,stat);
}

struct SQLStatement* prepareCall_OclDatasource(struct OclDatasource* self, char* stat)
{
  return prepare_OclDatasource(self,stat);
}


struct OclIterator* query_String_OclDatasource(struct OclDatasource* self, char* stat)
{ char* errs;
  char** res; 
  int nrow; 
  int ncol; 
  char** cnames; 

  if (self->database != NULL) 
  { int succ = sqlite3_get_table(self->database, stat, &res, &nrow, &ncol, &errs); 
    if (succ == SQLITE_OK)
    { struct ocltnode** resultSet = 
        (struct ocltnode**) calloc(nrow+1, sizeof(struct ocltnode*)); 

      cnames = (char**) calloc(ncol, sizeof(char*)); 

      int cind = 0; 
      for ( ; cind < ncol; cind++) 
      { cnames[cind] = res[cind]; } 
      cnames[cind] = NULL; 

      int r = 1; 
      for ( ; r <= nrow; r++) 
      { int c = 0;
        struct ocltnode* record = NULL; 
 
        for ( ; c < ncol; c++) 
        { int indx = r*ncol + c; 

          /* printf("%s |-> %s ", res[c], res[indx]); */ 

          record = insertIntoMap(record, res[c], res[indx]);  
        }    

        /* printf("\n"); */
 
        resultSet[r-1] = record; 
      } 
      resultSet[r-1] = NULL; 

      /* printf("There are %d records\n", length(resultSet)); */
 
      struct OclIterator* iter = newOclIterator_Sequence(resultSet); 
      iter->columnNames = cnames; 
      return iter;
    }  
  } 
  return NULL; 
}

struct OclIterator* rawQuery_OclDatasource(struct OclDatasource* self, char* stat, void* pos[])
{
  return query_String_OclDatasource(self,stat);
}

/*
public String nativeSQL(struct OclDatasource* self, String stat)
  {
    String result = null;
    return result;
  }
*/ 

struct OclIterator* query_String_Sequence_OclDatasource(struct OclDatasource* self, char* stat, char* cols[])
{
  return query_String_OclDatasource(self,stat);
}


void execSQL_OclDatasource(struct OclDatasource* self, char* stat)
{ char* errs; 

  if (self->database != NULL)
  { sqlite3_exec(self->database, 
      stat, 
      NULL, NULL, &errs);
  } 
}


void abort_OclDatasource(struct OclDatasource* self)
{ if (self->database != NULL)
  { sqlite3_interrupt(self->database); } 
}


void close_OclDatasource(struct OclDatasource* self)
{ if (self->database != NULL)
  { sqlite3_close(self->database); }
}


void commit_OclDatasource(struct OclDatasource* self)
{  }


void rollback_OclDatasource(struct OclDatasource* self)
{  }


void connect_OclDatasource(struct OclDatasource* self)
{  }


struct OclDatasource* openConnection_OclDatasource(void)
{ return NULL; }

void setSchema_OclDatasource(struct OclDatasource* self, char* s)
{
  self->schema = s;
}


char* getSchema_OclDatasource(struct OclDatasource* self)
{
  return self->schema;
}


struct OclFile* getInputStream_OclDatasource(struct OclDatasource* self)
{ return NULL; }


struct OclFile* getOutputStream_OclDatasource(struct OclDatasource* self)
{ return NULL; }


char* getURL_OclDatasource(struct OclDatasource* self)
{
  return self->url; 
}


void* getContent_OclDatasource(struct OclDatasource* self)
{ return NULL; }


char* getFile_OclDatasource(struct OclDatasource* self)
{
  return self->file; 
}


char* getHost_OclDatasource(struct OclDatasource* self)
{
  return self->host; 
}


int getPort_OclDatasource(struct OclDatasource* self)
{
  return self->port; 
}


char* getProtocol_OclDatasource(struct OclDatasource* self)
{
  return self->protocol;
}




struct SQLStatement* newSQLStatement(void) { 
  struct SQLStatement* res = (struct SQLStatement*) malloc(sizeof(struct SQLStatement)); 
  res->text = ""; 
  res->connection = NULL; 
  res->statement = NULL; 
  res->resultSet = NULL; 
  return res;
} 


void close_SQLStatement(struct SQLStatement* self)
{ if (self->statement != NULL) 
  { sqlite3_finalize(self->statement); }

  if (self->connection != NULL) 
  { close_OclDatasource(self->connection); } 
} 

void closeOnCompletion_SQLStatement(struct SQLStatement* self)
{ } 

void setString_SQLStatement(struct SQLStatement* self, int field, char* value)
{ if (self->statement != NULL) 
  { sqlite3_bind_text(self->statement, field, value, strlen(value), NULL); } 
} 

void setInt_SQLStatement(struct SQLStatement* self, int field, int value)
{ if (self->statement != NULL) 
  { sqlite3_bind_int(self->statement, field, value); } 
} 

void setByte_SQLStatement(struct SQLStatement* self, int field, int value)
{ if (self->statement != NULL) 
  { sqlite3_bind_int(self->statement, field, value); } 
} 

void setShort_SQLStatement(struct SQLStatement* self, int field, int value)
{ if (self->statement != NULL) 
  { sqlite3_bind_int(self->statement, field, value); } 
} 

void setBoolean_SQLStatement(struct SQLStatement* self, int field, unsigned char value)
{ if (self->statement != NULL) 
  { sqlite3_bind_int(self->statement, field, value); } 
} 

void setLong_SQLStatement(struct SQLStatement* self, int field, long value)
{ if (self->statement != NULL) 
  { sqlite3_bind_int64(self->statement, field, value); } 
} 

void setDouble_SQLStatement(struct SQLStatement* self, int field, double value)
{ if (self->statement != NULL) 
  { sqlite3_bind_double(self->statement, field, value); } 
} 

void setTimestamp_SQLStatement(struct SQLStatement* self, int field, struct OclDate* value)
{ if (self->statement != NULL) 
  { sqlite3_bind_int64(self->statement, field, value->time); } 
} 

void setNull_SQLStatement(struct SQLStatement* self, int field, void* value)
{ if (self->statement != NULL) 
  { sqlite3_bind_null(self->statement, field); } 
} 

 
struct OclIterator* executeQuery_SQLStatement(struct SQLStatement* self, char* stat)
{ if (self->connection != NULL) 
  { self->resultSet = query_String_OclDatasource(self->connection, stat); } 
  return self->resultSet; 
}

void execute_SQLStatement(struct SQLStatement* self, char* stat)
{ if (self->connection != NULL)
  { execSQL_OclDatasource(self->connection,stat); } 
}  

void executeUpdate_SQLStatement(struct SQLStatement* self)
{ if (self->statement != NULL) 
  { sqlite3_step(self->statement); } 
  else 
  { execute_SQLStatement(self, self->text); }  
}

struct OclDatasource* getConnection_SQLStatement(struct SQLStatement* self) 
{ return self->connection; } 

struct OclIterator* getResultSet_SQLStatement(struct SQLStatement* self) 
{ return self->resultSet; } 

void cancel_SQLStatement(struct SQLStatement* self) 
{ if (self->statement != NULL) 
  { sqlite3_reset(self->statement); } 
}


int main(int _argc, char* _argv[])
{ /* sqlite3* db;
  char* errs;  
  int succ = sqlite3_open("authors.db", &db); 
  printf("%d\n", succ); 
  if (succ == SQLITE_OK)
  { printf("Database opened correctly\n"); 
    succ = sqlite3_exec(db, 
      "create table Authors (authorId text, firstName text, surname text)", 
      NULL, NULL, &errs); 
    succ = sqlite3_exec(db, 
      "insert into Authors (authorId, firstName, surname) values ('1', 'Jim', 'Ballard')", 
      NULL, NULL, &errs);
    printf("%d\n", succ); 
    succ = sqlite3_exec(db, 
      "insert into Authors (authorId, firstName, surname) values ('2', 'Tom', 'Jones')", 
      NULL, NULL, &errs);
    printf("%d\n", succ); 

    char** res; 
    int nrow; 
    int ncol; 
    succ = sqlite3_get_table(db, "select * from Authors", &res, &nrow, &ncol, &errs); 
    printf("%d\n", succ); 
    printf("%d\n", nrow);
    printf("%d\n", ncol);

    struct ocltnode** resultSet = 
       (struct ocltnode**) calloc(nrow+1, sizeof(struct ocltnode*)); 

    int r = 1; 
    for ( ; r <= nrow; r++) 
    { int c = 0;
      struct ocltnode* record = NULL; 
 
      for ( ; c < ncol; c++) 
      { int indx = r*ncol + c; 

        printf("%s |-> %s ", res[c], res[indx]);
        record = insertIntoMap(record, res[c], res[indx]);  
      }    
      printf("\n"); 
      resultSet[r-1] = record; 
    } 
    resultSet[r-1] = NULL; 
    printf("There are %d records\n", length(resultSet)); 
    struct OclIterator* iter = newOclIterator_Sequence(resultSet); 
  } 
  else 
  { printf("ERROR\n"); } 
  sqlite3_close(db); */ 

  struct OclDatasource* ds = getConnection("authors.db", "", ""); 

 /* 
  execSQL_OclDatasource(ds, "insert into Authors (authorId, firstName, surname) values ('4', 'Sue', 'George')"); */ 

  struct OclIterator* iter = query_String_OclDatasource(ds, "select * from Authors"); 

  printf("%d\n", getColumnCount_OclIterator(iter)); 

  next_OclIterator(iter); 

  printf("%s\n", getCurrentFieldByIndex_OclIterator(iter,3)); 

  next_OclIterator(iter); 

  printf("%s\n", getCurrentFieldByIndex_OclIterator(iter,3)); 
 
  close_OclDatasource(ds);

  return 0; 
}

