package ocliterator

import "container/list"
import "ocl"
// import "fmt"

type OclIterator struct {
  position int
  markedPosition int
  elements *list.List
  columnNames *list.List
  currentElement *list.Element
  generatorFunction func(int) interface{}
}

type OclIteratorResult struct {
  done bool
  value interface{}
}

func createOclIteratorResult() *OclIteratorResult {
  var res *OclIteratorResult
  res = &OclIteratorResult{}
  return res
}

func NewOclIteratorResult(v interface{}) *OclIteratorResult {
  var result *OclIteratorResult
  result = createOclIteratorResult()
  if v == nil { 
    result.done = true
  } else { 
    result.done = false
  } 
  result.value = v
  return result
} 
  
func createOclIterator() *OclIterator {
  var res *OclIterator
  res = &OclIterator{}
  return res
}

func NewOclIterator_Sequence(col *list.List) *OclIterator {
  var result *OclIterator
  result = createOclIterator()
  result.elements = col
  result.columnNames = list.New()
  result.position = 0
  result.markedPosition = 0
  if col.Len() > 0 { 
    result.currentElement = col.Front()
  } else { 
    result.currentElement = nil
  } 

  return result
}

func NewOclIterator_Set(col *list.List) *OclIterator {
  var result *OclIterator
  result = createOclIterator()
  result.elements = col
  result.columnNames = list.New()
  result.position = 0
  result.markedPosition = 0
  if col.Len() > 0 { 
    result.currentElement = col.Front()
  } else { 
    result.currentElement = nil
  } 
  return result
}

func NewOclIterator_String(s string) *OclIterator {
  var result *OclIterator
  result = createOclIterator()
  col := ocl.Split(s,"[ \\n\\t\\r]+")
  result.elements = col
  result.columnNames = list.New()
  result.position = 0
  result.markedPosition = 0
  if col.Len() > 0 { 
    result.currentElement = col.Front()
  } else { 
    result.currentElement = nil
  } 
  return result
}

func NewOclIterator_String_String(s string, r string) *OclIterator {
  var result *OclIterator
  result = createOclIterator()
  col := ocl.Split(s,"[" + r + "]+")
  result.elements = col
  result.columnNames = list.New()
  result.position = 0
  result.markedPosition = 0
  if col.Len() > 0 { 
    result.currentElement = col.Front()
  } else { 
    result.currentElement = nil
  } 
  return result
}

func NewOclIterator_Function(f func(int) interface{}) *OclIterator {
  var result *OclIterator
  result = createOclIterator()
  result.elements = list.New()
  result.columnNames = list.New()
  result.position = 0
  result.markedPosition = 0
  result.generatorFunction = f
  return result
}

func (self *OclIterator) SetColumnNames(names *list.List) { 
  self.columnNames = names
} 

func (self *OclIterator) At(i int) interface{} { 
  return ocl.At(self.elements,i)
} 

func (self *OclIterator) GetPosition() int { 
  return self.position
} 

func (self *OclIterator) MarkPosition() { 
  self.markedPosition = self.position
} 

func (self *OclIterator) HasNext() bool { 
  return 0 <= self.position && 
         self.position < self.elements.Len()
}

func (self *OclIterator) HasPrevious() bool {
  return self.position > 1 &&
         self.position <= self.elements.Len() + 1
}

func (self *OclIterator) IsAfterLast() bool { 
  return self.position > self.elements.Len()
}

func (self *OclIterator) IsBeforeFirst() bool {
  return self.position < 1
}

func (self *OclIterator) NextIndex() int { 
  return self.position + 1
}

func (self *OclIterator) PreviousIndex() int {
  return self.position - 1
}

func (self *OclIterator) MoveForward() {
  if self.position == 0 { 
    self.currentElement = self.elements.Front()
  } else { 
    if self.currentElement != nil {
      self.currentElement = self.currentElement.Next() 
    }
  } 

  if self.position <= self.elements.Len() { 
    self.position = self.position + 1
  }
}

func (self *OclIterator) MoveBackward() {
  if self.position == self.elements.Len() + 1 { 
    self.currentElement = self.elements.Back()
  } else {
    if self.currentElement != nil {
      self.currentElement = self.currentElement.Prev() 
    }
  } 

  if self.position >= 1 {
    self.position = self.position - 1
  }
}

func (self *OclIterator) MoveTo(n int) { 
  if 0 <= n && n <= self.elements.Len() + 1 { 
    self.position = n
    i := 0
    for x := self.elements.Front(); x != nil; x = x.Next() { 
      if n == i+1 {
        self.currentElement = x
        return
      } 
      i = i+1
    }
  }
}

func (self *OclIterator) SetPosition(n int) { 
  self.MoveTo(n)
}

func (self *OclIterator) MoveToMarkedPosition() { 
  self.MoveTo(self.markedPosition)
}

func (self *OclIterator) MovePosition(i int) { 
  self.MoveTo(self.position + i)
}


func (self *OclIterator) MoveToFirst() { 
  self.MoveTo(1)
}

func (self *OclIterator) MoveToLast() { 
  self.MoveTo(self.elements.Len())
}

func (self *OclIterator) MoveToStart() { 
  self.MoveTo(0)
}

func (self *OclIterator) MoveToEnd() { 
  self.MoveTo(self.elements.Len() + 1)
}

func (self *OclIterator) GetCurrent() interface{} { 
  if self.currentElement != nil { 
    return self.currentElement.Value
  } 

  if 1 <= self.position && 
     self.position <= self.elements.Len() {
    return ocl.At(self.elements,self.position) 
  }

  return nil
}

func (self *OclIterator) getElement() *list.Element {
  i := 0
  for x := self.elements.Front(); x != nil; x = x.Next() { 
    if self.position == i+1 {
      return x
    } 
    i = i+1
  }
  return nil
}

func (self *OclIterator) Set(v interface{}) { 
  if self.currentElement != nil { 
    self.currentElement.Value = v
    return
  } 

  i := 0
  for x := self.elements.Front(); x != nil; x = x.Next() { 
    if self.position == i+1 {
      x.Value = v
      return
    } 
    i = i+1
  }
}

func (self *OclIterator) Remove() { 
  if self.currentElement != nil { 
    newCurrent := self.currentElement.Next()
    self.elements.Remove(self.currentElement)
    self.currentElement = newCurrent
    return
  } 

  x := self.getElement()
  if x != nil {
    self.elements.Remove(x)
  }
}

func (self *OclIterator) Insert(v interface{}) { 
  if self.currentElement != nil { 
    self.currentElement = self.elements.InsertBefore(v,self.currentElement)
    return
  }

  x := self.getElement()
  if x != nil {
    self.elements.InsertBefore(v,x)
  }
}
  

func (self *OclIterator) Next() interface{} {
  self.MoveForward()
  curr := self.GetCurrent()
  return curr 
} 

func (self *OclIterator) NextResult() *OclIteratorResult {
  if self.generatorFunction != nil { 
    r := self.generatorFunction(self.position)
    self.position = self.position + 1
    if self.position <= self.elements.Len() { 
      self.Set(r)
    } else { 
      self.elements.PushBack(r)
    }
    return NewOclIteratorResult(r) 
  }
  self.MoveForward()
  curr := self.GetCurrent()
  return NewOclIteratorResult(curr) 
} 

func (self *OclIterator) Previous() interface{} {
  self.MoveBackward()
  curr := self.GetCurrent()
  return curr
} 

func (self *OclIterator) Length() int { 
  return self.elements.Len()
} 


func (self *OclIterator) Close() { 
  self.position = 0
  self.markedPosition = 0
  self.elements = nil
  self.columnNames = nil
  self.currentElement = nil
}

func (self *OclIterator) GetColumnCount() int { 
  return self.columnNames.Len()
} 

func (self *OclIterator) GetColumnName(i int) string {
  res := ocl.At(self.columnNames,i)
  return res.(string)
}

func (self *OclIterator) GetCurrentFieldByIndex(i int) interface{} { 
  fld := self.GetColumnName(i)
  curr := self.GetCurrent()
  var m map[string]interface{}
  m = curr.(map[string]interface{})
  return m[fld]
}

func (self *OclIterator) SetCurrentFieldByIndex(i int, v interface{}) { 
  fld := self.GetColumnName(i)
  curr := self.GetCurrent()
  var m map[string]interface{}
  m = curr.(map[string]interface{})
  m[fld] = v
}




